﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.BioBoost.ProblemDescription;
using HeuristicLab.BioBoost.Representation;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.BioBoost.Operators.Mutation {

  public class PlantSplitter : SingleSuccessorOperator, IManipulator, IStochasticOperator {

    #region Parameters
    public LookupParameter<BioBoostProblemData> ProblemDataParameter { get { return (LookupParameter<BioBoostProblemData>) Parameters["ProblemData"]; } }
    public LookupParameter<DistanceMatrix> DistanceMatrixParameter { get { return (LookupParameter<DistanceMatrix>) Parameters["DistanceMatrix"]; } }
    public ValueParameter<BoolValue> DistanceMatrixInProblemDataParameter { get { return (ValueParameter<BoolValue>) Parameters["DistanceMatrixInProblemData"]; } }
    public ValueLookupParameter<IntMatrix> RegionBoundsParameter { get { return (ValueLookupParameter<IntMatrix>) Parameters["RegionBounds"]; } } 
    public ILookupParameter<IRandom> RandomParameter { get { return (ILookupParameter<IRandom>) Parameters["Random"]; } }
    public IValueLookupParameter<PercentValue> MergeAvoidanceProbabilityParameter { get { return (IValueLookupParameter<PercentValue>) Parameters["MergeAvoidanceProbability"]; } }
    public IValueLookupParameter<PercentValue> SplitProbabilityParameter { get { return (IValueLookupParameter<PercentValue>) Parameters["SplitProbability"]; } }
    public IValueLookupParameter<PercentValue> MovePercentageParameter { get { return (IValueLookupParameter<PercentValue>) Parameters["MovePercentage"]; } }
    #endregion

    #region Parameter Values
    public BioBoostProblemData ProblemData { get { return ProblemDataParameter.ActualValue; } }
    public IntMatrix RegionBounds { get { return RegionBoundsParameter.ActualValue; } }
    public IRandom Random { get { return RandomParameter.ActualValue; } }
    public bool DistanceMatrixInProblemData { get { return DistanceMatrixInProblemDataParameter.Value.Value; } }
    public DistanceMatrix DistanceMatrix {
      get {
        if (DistanceMatrixInProblemData) {
          return ((IValueParameter<DistanceMatrix>) ProblemData.Parameters[DistanceMatrixParameter.ActualName]).Value;
        } else {
          return DistanceMatrixParameter.ActualValue;
        }
      }
    }
    public double MergeAvoidanceProbability { get { return MergeAvoidanceProbabilityParameter.Value.Value; } }
    public double SplitProbability { get { return SplitProbabilityParameter.Value.Value; } }
    public double MovePercentage { get { return MovePercentageParameter.Value.Value; } }
    #endregion

    #region Construction & Cloning
    [StorableConstructor]
    public PlantSplitter(bool isDeserializing) {}
    public PlantSplitter(PlantSplitter orig, Cloner cloner) : base(orig, cloner) {}

    public PlantSplitter() {
      Parameters.Add(new LookupParameter<BioBoostProblemData>("ProblemData", "The data store of the detailed problem description."));
      Parameters.Add(new LookupParameter<DistanceMatrix>("DistanceMatrix", "The distance matrix to use", "StreetDistanceMatrix")); // TODO: check this
      Parameters.Add(new ValueParameter<BoolValue>("DistanceMatrixInProblemData", "Whether to look for the distance matrix in problem data or in scope.", new BoolValue(true)));
      Parameters.Add(new ValueLookupParameter<IntMatrix>("RegionBounds", "The limits of valid region ids."));
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator."));
      Parameters.Add(new ValueLookupParameter<PercentValue>("MergeAvoidanceProbability", "Probability that a moved plant will avoid existing plant locations.", new PercentValue(1)));
      Parameters.Add(new ValueLookupParameter<PercentValue>("SplitProbability", "Probability that only part of the suppliers will follow to the new location.", new PercentValue(1)));
      Parameters.Add(new ValueLookupParameter<PercentValue>("MovePercentage", "In case a plant is split, the percentage of suppliers that will move.", new PercentValue(0.5)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new PlantSplitter(this, cloner);
    }
    #endregion

    public override IOperation Apply() {
      var scope = ExecutionContext.Scope;
      var dm = DistanceMatrix;
      var feedstock =
        ProblemData.TransportTargets.CheckedItems.ElementAt(
          Random.Next(ProblemData.TransportTargets.CheckedItems.Count())).Value.Value;
      string product = null;
      var productLinks = ProblemData.ProductLinks;
      for (int i = 0; i < productLinks.Rows; i++) {
        if (productLinks[i, 0] == feedstock)
          product = productLinks[i, 1];
      }
      var supplyTransports = scope.Variables[LayerDescriptor.TransportTargets.NameWithPrefix(feedstock)].Value as IntegerVector;
      var productTransportName = LayerDescriptor.TransportTargets.NameWithPrefix(product);
      var productTransports = product == null || !scope.Variables.ContainsKey(productTransportName) ? null
        : scope.Variables[productTransportName].Value as IntegerVector;
      if (supplyTransports != null) {
        var plants = supplyTransports.Select((target, source) => new {target, source}).GroupBy(p => p.target).ToList();
        var plant = plants.ElementAt(Random.Next(plants.Count));
        var forbiddenRegions = new HashSet<int>();
        if (Random.NextDouble() > MergeAvoidanceProbability) {
          forbiddenRegions = new HashSet<int>(plants.Select(p => p.Key));
        }
        var newTarget = FindNewTarget(plant.Key, Random, dm, forbiddenRegions);
        bool split = Random.NextDouble() < SplitProbability;
        foreach (var supplier in plant) {
          if (!split || Random.NextDouble() <= MovePercentage) {
            supplyTransports[supplier.source] = newTarget;
          }
        }
        if (productTransports != null) {
          var temp = productTransports[newTarget];
          productTransports[newTarget] = productTransports[plant.Key];
          productTransports[plant.Key] = temp;
        }
      }
      return base.Apply();
    }

    private int FindNewTarget(int oldTarget, IRandom random, DistanceMatrix distances, HashSet<int> forbiddenRegions) {
      var neighborDistances = new double[0].Select((d, idx) => new {index = idx, distance = d}).ToList();
      var maxDistance = 0d;
      for (int j = 0; j < distances.Columns; j++) {
        if (forbiddenRegions.Contains(j)) continue;
        var dist = distances[oldTarget, j];
        neighborDistances.Add(new {index = j, distance = dist});
        maxDistance = Math.Max(dist, maxDistance);
      }
      neighborDistances = neighborDistances.Select(p => new {p.index, distance = maxDistance - p.distance}).ToList();
      neighborDistances.Sort((a, b) => b.distance.CompareTo(a.distance));
      var totalDistance = neighborDistances.Sum(p => p.distance);
      var threshold = random.NextDouble()*totalDistance;
      var sum = 0d;
      var index = 0;
      while (index < neighborDistances.Count && sum < threshold) {
        sum += neighborDistances[index].distance;
        index++;
      }
      index = Math.Min(index, neighborDistances.Count - 1);
      return neighborDistances[index].index;
    }

  }
}
